#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright 2019, Nils Hilbricht, Germany ( https://www.hilbricht.net )

This file is part of the Laborejo Software Suite ( https://www.laborejo.org ),
more specifically its template base application.

The Template Base Application is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging; logging.info("import {}".format(__file__))

#Standard Library Modules
import os.path
#Third Party Modules
from PyQt5 import QtWidgets, QtCore, QtGui
#Template Modules
from template.qtgui.mainwindow import MainWindow as TemplateMainWindow
from template.qtgui.menu import Menu
from template.qtgui.about import About
from template.qtgui.flowlayout import FlowLayout
#Client modules
import engine.api as api
from .designer.channelstrip import Ui_ChannelStrip

class MainWindow(TemplateMainWindow):
    
    def __init__(self): 
        """The order of calls is very important. 
        THe split ploint is calling the super.__init. Some functions need to be called before,
        some after.
        For example:
        
        The about dialog is created in the template main window init. So we need to set additional
        help texts before that init.
       
        The menu needs the template main window to exist
        so it can reference the ui.menuActions.        
        """     
        
        #Inject more help texts in the templates About "Did You Know" field.
        ############################
        #About.didYouKnow is a class variable.
        #Make the first three words matter!
        #Do not start them all with "You can..." or "...that you can", in response to the Did you know? title.        
        About.didYouKnow = [
            QtCore.QCoreApplication.translate("About", "Block incoming midi instrument changes by checking 'Ignore MIDI Bank and Progam Changes' in the main window. This will make your carefully selected setup very reliable."),
            QtCore.QCoreApplication.translate("About", "A short flash of color indicates which channel just received a note."),
            QtCore.QCoreApplication.translate("About", "Click with middle mouse button on a channel to hear a short test melody with the chosen instrument."),
            ] + About.didYouKnow                
        
        #Init the templates main window 
        ############################
        super().__init__()      
        
        
        #Set up the custom interface
        ############################        
        self.highlightCyanPalette = QtGui.QPalette(self.fPalBlue)                        
        self.highlightCyanPalette.setColor(QtGui.QPalette.Disabled, QtGui.QPalette.Window, QtGui.QColor("cyan"))
        self.highlightCyanPalette.setColor(QtGui.QPalette.Active,   QtGui.QPalette.Window, QtGui.QColor("cyan"))
        self.highlightCyanPalette.setColor(QtGui.QPalette.Inactive, QtGui.QPalette.Window, QtGui.QColor("cyan"))


        self.ui.label_soundfont_name.hide() #will be shown after loading an sf2
        
        self.ui.checkBox_ignoreProgramChanges.hide() #will be shown after loading an sf2
        self.ui.checkBox_ignoreProgramChanges.stateChanged.connect(api.setIgnoreProgramAndBankChanges) #one parameter: int state. 0 is unchecked, 1 is partially, 2 is checked
        
        self.rackFlowLayout = FlowLayout(margin=20, spacing=20)
        self.ui.scrollAreaWidgetContents.setLayout(self.rackFlowLayout)
        self.channels = []
        for i in range(16):
            cs = ChannelStrip(mainWindow = self, number=i+1) #Channels are 1 based
            self.channels.append(cs)
            cs.hide()
            self.rackFlowLayout.addWidget(cs)
        
        
        #Api Callbacks
        #Must be registered before startEngine, which is in super.__init__
        ############################        
        api.callbacks.soundfontChanged.append(self.callback_SoundFontLoaded)
        api.callbacks.channelChanged.append(self.callback_channelChanged)
        api.callbacks.channelActivity.append(self.callback_channelActivity)
        api.callbacks.ignoreProgramChangesChanged.append(self.callback_ignoreProgramChangesChanged)   

        #New menu entries and template-menu overrides
        self.menu.connectMenuEntry("actionOpen_sf2_Soundfont", self.openSoundfontDialog) 
        #self.menu.addMenuEntry("menuEdit", "actionNils", "Nils", lambda: print("Merle"))
        #self.menu.connectMenuEntry("actionNils", lambda: print("Perle"))        
       
        self.ui.actionUndo.setVisible(False)
        self.ui.actionRedo.setVisible(False)
        self.ui.menuEdit.menuAction().setVisible(False) 
      
       
        self.start()  #Starts the engine, starts the eventLoop, shows the GUI, or not, depends on the NSM gui save setting. We need to call that after the menu, otherwise the about dialog will block and then we get new menu entries, which looks strange.

        
    def openSoundfontDialog(self):
        """Called by the menu and GUI directly to choose a soundfont file.
        We forward the path to NSM which gives us a link in our session dir which we 
        give to the api. 
        The api will react with a callback and we can show our Channel Strips"""                        
        if "lastOpenDirectory" in api.session.guiSharedDataToSave:                    
            defaultDir = api.session.guiSharedDataToSave["lastOpenDirectory"]
        else:
            defaultDir = os.path.expanduser("~")
                        
        filePath, extension = QtWidgets.QFileDialog.getOpenFileName(self, "Open Soundfont", defaultDir, "Soundfont 2 (*.sf2);;All Files (*.*)")        
        if filePath: #not cancelled                        
            linkedPath = self.nsmClient.importResource(filePath)            
            api.session.guiSharedDataToSave["lastOpenDirectory"] = os.path.dirname(filePath)
            success = api.loadSoundfont(linkedPath)
            if not success:
                for cs in self.channels:
                    self.ui.label_soundfont_name.hide() #will be shown after loading an sf2        
                    self.ui.checkBox_ignoreProgramChanges.hide() #will be shown after loading an sf2
                    cs.hide()

    def callback_ignoreProgramChangesChanged(self, state):
        """Redundant while the program runs. Useful for load / startup"""        
        self.ui.checkBox_ignoreProgramChanges.blockSignals(True)
        self.ui.checkBox_ignoreProgramChanges.setChecked(state)
        self.ui.checkBox_ignoreProgramChanges.blockSignals(False)
        
    def updateChannelStrips(self, exportDict):
        for channelStrip in self.channels:             
            channelStrip.update(exportDict)
            channelStrip.show()
    
    def callback_SoundFontLoaded(self, exportDict):
        """Can be called multiple times. On load, first time file open, other file open"""
        self.updateChannelStrips(exportDict)
        self.setWindowTitle(exportDict["name"])
        self.ui.label_soundfont_name.setText(exportDict["name"])
        self.ui.label_soundfont_name.show()
        self.ui.checkBox_ignoreProgramChanges.show()
    
    def callback_channelChanged(self, channel, exportDict):
        """A single channel changed bank, program or both"""        
        self.channels[channel-1].callback_channelChanged(exportDict) #self.channels as python list is 0 based, channels are 1 based

    def callback_channelActivity(self, channel):
        self.channels[channel-1].activity() #self.channels as python list is 0 based, channels are 1 based

    def dropEvent(self, event):
        """This function does not exist in the template.
        It is easiest to edit it directly than to create another abstraction layer.
                
        Having that function in the mainWindow will not make drops available for subwindows
        like About or UserManual. """        
        for url in event.mimeData().urls():
            filePath = url.toLocalFile()
            #Decide here if you want only files, only directories, both etc.            
            if os.path.isfile(filePath) and filePath.lower().endswith(".sf2"):                
                linkedPath = self.nsmClient.importResource(filePath)                            
                api.loadSoundfont(linkedPath)



class ChannelStrip(QtWidgets.QGroupBox):
    """Created once on program start. From then on only updated values"""    

    def __init__(self, mainWindow, number):        
        super().__init__()
        self.mainWindow = mainWindow
        self.ui =  Ui_ChannelStrip()
        self.ui.setupUi(self)                            
        self.channelNumber = number
        self.setTitle(f"Channel {self.channelNumber}")
        
        if number % 2:            
            self.setAutoFillBackground(True) #otherwise the palette doesn't work
            self.setBackgroundRole(mainWindow.fPalBlue.AlternateBase)               
                 
        self.ui.comboBox_bank.currentIndexChanged.connect(self.reactToBankOrProgramChange)
        self.ui.comboBox_program.currentIndexChanged.connect(self.reactToBankOrProgramChange)               
        
        self.activeFlag = False
        api.session.eventLoop.slowConnect(self._activityOff)
            
    def _activityOff(self):        
        if self.activeFlag:
            self.activeFlag = False
            self.setPalette(self.mainWindow.fPalBlue)
    
    def activity(self):
        """Show midi note ons as flashing light.
        turning off is done by a 60ms timer for all channels."""                        
        self.setPalette(self.mainWindow.highlightCyanPalette)
        self.activeFlag = True    
        #QtCore.QTimer().singleShot(50, self._activityOff)         this doesn't scale too well
        
    def _populateComboBox_Program(self, bank, program):
        self.ui.comboBox_program.clear()        
        assert self.patchlist[bank] #not empty.
        assert 0 <=  program < 128, program
        nowProgram =  self.ui.comboBox_program.currentIndex()  #if below fails at least we don't get an error. Worst case is the wrong label.
        for prgnr, patchname in self.patchlist[bank].items():
            #Not every program is in every bank. But we don' want a drop down list with gaps.
            #So we save the program number as user data, which is only possile with addItem, not list based addItems
            text = str(prgnr).zfill(3) + " " + patchname                        
            self.ui.comboBox_program.addItem(text, prgnr) #Icon, Text, UserData
            if prgnr == program:
                nowProgram = self.ui.comboBox_program.count()-1 
                
        self.ui.comboBox_program.setCurrentIndex(nowProgram) #both 0 based

        
    def update(self, exportDict):
        """Fill in new data for a new soundfont.
        The engine guarantees that the activePatches are actually in the patchlist.
        This is important for a soundfont change to one that has fewer instruments.
        The engine handles this reduction."""
                
        self.ui.comboBox_bank.blockSignals(True)
        self.ui.comboBox_program.blockSignals(True)
        
        self.exportDict = exportDict
        bank, program, name = exportDict["activePatches"][self.channelNumber]  #fluidsynth-channels from 1, gui channels from 1                    
        self.setTitle(f"Channel {self.channelNumber}")
        self.patchlist = exportDict["patchlist"] #all patches in this soundfont. bank:{program:name}     

        #Banks
        self.ui.comboBox_bank.clear()
        for banknr in self.patchlist:            
            self.ui.comboBox_bank.addItem(str(banknr), banknr) #Icon, Text, UserData
            if banknr == bank:
                nowBank = self.ui.comboBox_bank.count()-1 
        self.ui.comboBox_bank.setCurrentIndex(nowBank)        

        #Programs
        self._populateComboBox_Program(bank, program)
        
        
        #Unblock Signals
        self.ui.comboBox_bank.blockSignals(False)
        self.ui.comboBox_program.blockSignals(False)      

    def reactToBankOrProgramChange(self, newIndex): #discard newIndex      
        bank = self.ui.comboBox_bank.currentData()
        program = self.ui.comboBox_program.currentData()
        assert not bank is None, bank
        assert not program is None, program
        api.setPatch(self.channelNumber, bank, program) # triggers self.callback_channelChanged

    def callback_channelChanged(self, exportDict):
        """Either triggered by a GUI action, routed through the api, or by incoming midi message"""
        assert self.channelNumber == exportDict["channel"], (self.channelNumber, exportDict["channel"])        
        self.ui.comboBox_bank.blockSignals(True)
        self.ui.comboBox_program.blockSignals(True)     
        
        comboIndex = self.ui.comboBox_bank.findData(exportDict["bank"])
        if comboIndex == -1: #qt not found
            raise ValueError("Bank not in patchlist")            
        else:
            self.ui.comboBox_bank.setCurrentIndex(comboIndex) #both 0 based
        
        self._populateComboBox_Program(exportDict["bank"], exportDict["program"])
                
        self.ui.comboBox_bank.blockSignals(False)        
        self.ui.comboBox_program.blockSignals(False)     

    def mousePressEvent(self, event):        
        if event.button() == QtCore.Qt.MiddleButton:
            api.playTestSignal(self.channelNumber)
            event.accept()
        else:
            event.ignore()
            super().mousePressEvent(event)

